class Network:
    BSC = "bsc"
    ETH = "eth"
    ATC = "atc"
    MATIC = "matic"
    TRX = "trx"
    XBN = "xbn"
    BASE = "base"
    
class AssetType: 
    FIAT = "fiat"
    WRAPPED = "wrapped" 
    ENAIRA = "enaira"



class ProviderType:
    KORAPAY = "korapay"
    BUDPAY = "budpay"
    # Add other provider types as needed



from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
import os
import base64
class AESCrypto:
    ALGORITHM = algorithms.AES
    IV_LENGTH = 16
    KEY_LENGTH = 32  # 256 bits
    BLOCK_SIZE = 16  # AES block size
    class InvalidPaddingError(Exception):
        pass
    @staticmethod
    def prepare_key(key: str) -> bytes:
        # Hash the key using SHA-256 to ensure it's always the correct length (32 bytes)
        digest = hashes.Hash(hashes.SHA256(), backend=default_backend())
        digest.update(key.encode('utf-8'))
        return digest.finalize()
    @staticmethod
    def pkcs7_pad(data: bytes) -> bytes:
        """Apply PKCS7 padding to the data."""
        padding_length = AESCrypto.BLOCK_SIZE - (len(data) % AESCrypto.BLOCK_SIZE)
        padding = bytes([padding_length] * padding_length)
        return data + padding
    @staticmethod
    def pkcs7_unpad(data: bytes) -> bytes:
        """
        Remove and validate PKCS7 padding.
        Raises InvalidPaddingError if padding is incorrect.
        """
        if len(data) == 0:
            raise AESCrypto.InvalidPaddingError("Empty data")
        if len(data) % AESCrypto.BLOCK_SIZE != 0:
            raise AESCrypto.InvalidPaddingError("Data length not multiple of block size")
        padding_length = data[-1]
        if padding_length == 0 or padding_length > AESCrypto.BLOCK_SIZE:
            raise AESCrypto.InvalidPaddingError("Invalid padding length")
        if len(data) < padding_length:
            raise AESCrypto.InvalidPaddingError("Padding length larger than data")
        # Check all padding bytes are correct
        padding = data[-padding_length:]
        if not all(x == padding_length for x in padding):
            raise AESCrypto.InvalidPaddingError("Invalid padding values")
        return data[:-padding_length]
    @staticmethod
    def encrypt(data: str, key: str) -> dict:
        # Generate a random Initialization Vector (IV)
        iv = os.urandom(AESCrypto.IV_LENGTH)
        key_buffer = AESCrypto.prepare_key(key)
        # Create cipher and encrypt the data
        cipher = Cipher(AESCrypto.ALGORITHM(key_buffer), modes.CBC(iv), backend=default_backend())
        encryptor = cipher.encryptor()
        # Properly pad data using PKCS7
        padded_data = AESCrypto.pkcs7_pad(data.encode('utf-8'))
        encrypted = encryptor.update(padded_data) + encryptor.finalize()
        # Return the encrypted content and the IV (both base64 encoded)
        return {
            'content': base64.b64encode(encrypted).decode('utf-8'),
            'iv': base64.b64encode(iv).decode('utf-8')
        }
    @staticmethod
    def decrypt(encrypted_data: dict, key: str) -> str:
        try:
            # Decode the base64 encoded IV and content
            iv = base64.b64decode(encrypted_data['iv'])
            encrypted_content = base64.b64decode(encrypted_data['content'])
            key_buffer = AESCrypto.prepare_key(key)
            # Create cipher and decrypt the data
            cipher = Cipher(AESCrypto.ALGORITHM(key_buffer), modes.CBC(iv), backend=default_backend())
            decryptor = cipher.decryptor()
            decrypted = decryptor.update(encrypted_content) + decryptor.finalize()
            # Remove padding with validation
            unpadded = AESCrypto.pkcs7_unpad(decrypted)
            return unpadded.decode('utf-8')
        except AESCrypto.InvalidPaddingError as e:
            # Handle padding errors uniformly to prevent timing attacks
            raise AESCrypto.InvalidPaddingError("Decryption failed")
        except Exception as e:
            # Handle all other errors uniformly
            raise AESCrypto.InvalidPaddingError("Decryption failed")


from .CryptoWallet import CryptoWallet

class WalletManager:

    def generate_wallet_address(self, network: str) -> str:
        response =  CryptoWallet.generate_wallet_with_mnemonic_details(network)
        return {
            "success": True,
            "data": response
        }

    def validate_address(self, address, network):
        return CryptoWallet.validate_address(address, network)



#   ____ _   _  ____       __  __                                   
#  / ___| \ | |/ ___|_ __ |  \/  | __ _ _ __   __ _  __ _  ___ _ __ 
# | |   |  \| | |  _| '_ \| |\/| |/ _` | '_ \ / _` |/ _` |/ _ \ '__|
# | |___| |\  | |_| | | | | |  | | (_| | | | | (_| | (_| |  __/ |   
#  \____|_| \_|\____|_| |_|_|  |_|\__,_|_| |_|\__,_|\__, |\___|_|   
#                                                   |___/           
# 


import json
from typing import Optional, Dict, Any
import requests
from requests.exceptions import RequestException, HTTPError
from .AESCrypto import AESCrypto
from .Ed25519Crypto import Ed25519Crypto

"""
    CNGnManager class is a wrapper around the CNGn API.
    It provides methods to interact with the CNGn API endpoints.
    It uses the requests library to make HTTP requests to the API.
    It uses the AESCrypto and Ed25519Crypto classes to encrypt and decrypt data.
    It handles API errors and returns appropriate error messages.
    It returns JSON responses from the API.
"""

class CNGnManager:
    API_URL = "https://api.cngn.co"
    API_CURRENT_VERSION = "v1"

    def __init__(self, api_key: str, private_key: str, encryption_key: str):
        self.api_key = api_key
        self.api_url = self.API_URL
        self.private_key = private_key
        self.encryption_key = encryption_key
        self.client = requests.Session()
        self.client.headers.update({
            'Authorization': f'Bearer {self.api_key}',
            'Content-Type': 'application/json',
            'Accept': 'application/json',
        })

    def __make_calls(self, method: str, endpoint: str, data: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
        aes_crypto = AESCrypto()
        ed_crypto = Ed25519Crypto()

        try:
            url = f'{self.api_url}/{self.API_CURRENT_VERSION}/api{endpoint}'
            request_data = self._prepare_request_data(data, aes_crypto)
            response = self._send_request(method, url, request_data)
            return self._process_response(response, ed_crypto)

        except (RequestException, HTTPError) as e:
            return self._handle_request_error(e)
        except Exception as e:
            return self._handle_unexpected_error(e)

    def _prepare_request_data(self, data: Optional[Dict[str, Any]], aes_crypto: AESCrypto) -> Optional[str]:
        if data is None:
            return None
        json_data = json.dumps(data)
        return aes_crypto.encrypt(json_data, self.encryption_key)

    def _send_request(self, method: str, url: str, data: Optional[str]) -> requests.Response:
        return self.client.request(method, url, json=data)

    def _process_response(self, response: requests.Response, ed_crypto: Ed25519Crypto) -> Dict[str, Any]:
        response_data = response.json()
        if "data" in response_data:
            decrypted_response = ed_crypto.decrypt_with_private_key(self.private_key, response_data["data"])
            response_data["data"] = json.loads(decrypted_response)
        return response_data

    def _handle_request_error(self, error: RequestException) -> Dict[str, Any]:
        if error.response:
            return error.response.json()
        return {
            'success': False,
            'error': 'API request failed',
            'message': str(error),
            'status_code': error.response.status_code if error.response else None,
        }

    def _handle_unexpected_error(self, error: Exception) -> Dict[str, Any]:
        return {
            'success': False,
            'error': 'An unexpected error occurred',
            'message': str(error),
            'status_code': 500,
        }
    
    def get_balance(self) -> str:
        return self.__make_calls("GET", "/balance")

    def get_transaction_history(self, page: int = 1, limit: int = 10 ) -> str:
        return self.__make_calls("GET", f"/transactions?page{page}&limit={limit}")

    def withdraw(self, data: dict) -> str:
        return self.__make_calls("POST", "/withdraw", data)

    def redeem_assets(self, data: dict) -> str:
        return self.__make_calls("POST", "/redeemAsset", data)

    def create_virtual_account(self, data: dict) -> str:
        return self.__make_calls("POST", "/createVirtualAccount", data)

    def update_external_accounts(self, data: dict) -> str:
        return self.__make_calls("POST", "/updateBusiness", data)

    def get_banks(self):
        return self.__make_calls("GET", "/banks")
    
    def swap_asset(self, data: dict) -> str:
        return self.__make_calls("POST", "/swap", data)
    



from ecdsa import SigningKey, SECP256k1
from mnemonic import Mnemonic
from bip32utils import BIP32Key, BIP32_HARDEN
from tronpy import Tron
from tronpy.keys import PrivateKey, PublicKey
from hashlib import sha3_256
from stellar_sdk import Keypair, StrKey
import coinaddrvalidator
import nacl.signing
import nacl.encoding
from .constants import Network  # Assuming you have a Network class or Enum in constants

class CryptoWallet:

    DERIVATION_PATHS = {
        Network.ETH: "m/44'/60'/0'/0/0",
        Network.BSC: "m/44'/60'/0'/0/0",
        Network.ATC: "m/44'/60'/0'/0/0",
        Network.MATIC: "m/44'/60'/0'/0/0",
        Network.TRX: "m/44'/195'/0'/0/0",  # TRON's derivation path
        Network.XBN: "m/44'/703'/0'/0"     # XBN's derivation path
    }

    @staticmethod
    def generate_wallet_with_mnemonic_details(network: str):
        mnemo = Mnemonic("english")
        mnemonic = mnemo.generate(strength=128)
        return CryptoWallet.generate_wallet_from_mnemonic(mnemonic, network)

    @staticmethod
    def generate_wallet_from_mnemonic(mnemonic: str, network: str):
        if network == Network.XBN:
            return CryptoWallet.generate_xbn_wallet(mnemonic)
        elif network == Network.TRX:
            return CryptoWallet.generate_trx_wallet(mnemonic)

        private_key = CryptoWallet.get_private_key_from_mnemonic(mnemonic, network)
        public_key = CryptoWallet.get_public_key(private_key, network)
        address = CryptoWallet.get_address_from_public_key(public_key, network)
        
        return {'mnemonic': mnemonic, 'privateKey': private_key, 'address': address, 'network': network}

    @staticmethod
    def get_public_key(private_key: str, network: str):
        if network == Network.XBN:
            seed = nacl.signing.SigningKey(bytes.fromhex(private_key))
            return seed.verify_key.encode().hex()
        sk = SigningKey.from_string(bytes.fromhex(private_key), curve=SECP256k1)
        return sk.verifying_key.to_string().hex()

    @staticmethod
    def get_private_key_from_mnemonic(mnemonic: str, network: str):
        seed = Mnemonic.to_seed(mnemonic)
        derivation_path = CryptoWallet.DERIVATION_PATHS[network]
        # Use BIP32Key derivation according to the provided network path
        key = BIP32Key.fromEntropy(seed)
        for index in derivation_path.split("/")[1:]:
            if index.endswith("'"):
                index = int(index[:-1]) + 0x80000000
            else:
                index = int(index)
            key = key.ChildKey(index)
        return key.PrivateKey().hex()

    @staticmethod
    def get_address_from_public_key(public_key: str, network: str):
        if network in [Network.ETH, Network.BSC, Network.MATIC, Network.ATC]:
            return CryptoWallet.get_ethereum_style_address(public_key)
        elif network == Network.TRX:
            return CryptoWallet.get_tron_address_from_public_key(public_key)
        raise ValueError(f'Unsupported network: {network}')

    @staticmethod
    def get_ethereum_style_address(public_key: str):
        clean_public_key = public_key[2:] if public_key.startswith('04') else public_key
        hash = sha3_256(bytes.fromhex(clean_public_key)).digest()
        return '0x' + hash[-20:].hex()

    @staticmethod
    def generate_trx_wallet(mnemonic: str):
        private_key = CryptoWallet.get_private_key_from_mnemonic(mnemonic, Network.TRX)
        tron_private_key = PrivateKey(bytes.fromhex(private_key))
        address = tron_private_key.public_key.to_base58check_address()

        return {
            'mnemonic': mnemonic,
            'privateKey': tron_private_key.hex(),
            'address': address,
            'network': Network.TRX
        }

    @staticmethod
    def get_tron_address_from_public_key(public_key: str):
        public_key_bytes = bytes.fromhex(public_key)
        if len(public_key_bytes) != 64:
            raise ValueError("Invalid public key length")

        tron_address = PublicKey(public_key_bytes).to_base58check_address()
        return tron_address
    
    @staticmethod
    def generate_xbn_wallet(mnemonic: str):
        # Derive BIP32 seed from mnemonic
        seed = Mnemonic.to_seed(mnemonic)
        derivation_path = CryptoWallet.DERIVATION_PATHS[Network.XBN]
        master_key = BIP32Key.fromEntropy(seed)
        
        for index in derivation_path.split("/")[1:]:
            if index.endswith("'"):
                index = int(index[:-1]) + BIP32_HARDEN # Hardened key
            else:
                index = int(index)
            key = master_key.ChildKey(index)

        raw_secret_key = key.PrivateKey()
        secret_key = StrKey.encode_ed25519_secret_seed(raw_secret_key)
        keypair = Keypair.from_secret(secret_key)

        return {
            'mnemonic': mnemonic,
            'privateKey': keypair.secret,
            'address': keypair.public_key,
            'network': Network.XBN
        }

    @staticmethod
    def validate_address(address: str, network: str):
        if network in [Network.ETH, Network.BSC, Network.MATIC, Network.ATC]:
            result = coinaddrvalidator.validate(network, address)
            print(f'Validation result for {network}: {result}')
            return result
        elif network == Network.TRX:
            tron = Tron()
            return tron.is_address(address)
        elif network == Network.XBN:
            result = coinaddrvalidator.validate(network, address)
            print(f'Validation result for {network}: {result}')
            return result
        raise ValueError(f'Unsupported network: {network}')


from nacl.public import PrivateKey, PublicKey, Box
from nacl.encoding import Base64Encoder
from nacl.bindings import crypto_sign_ed25519_sk_to_curve25519
import base64
import re

class Ed25519Crypto:
    is_initialized = False

    @staticmethod
    def initialize():
        if not Ed25519Crypto.is_initialized:
            # No explicit initialization is required for PyNaCl, but we keep this as a safeguard.
            Ed25519Crypto.is_initialized = True

    @staticmethod
    def parse_openssh_private_key(private_key: str) -> bytes:
        """
        Parses an OpenSSH Ed25519 private key to extract the 32-byte private key.
        
        :param private_key: The OpenSSH private key string.
        :return: The 32-byte Ed25519 private key as bytes.
        """
        # Remove the key header/footer and decode the base64 content
        private_key_stripped = re.sub(r'-----.* PRIVATE KEY-----', '', private_key).strip()
        private_key_stripped = re.sub(r"\s+", '', private_key_stripped)
        private_key_buffer = base64.b64decode(private_key_stripped)

        # Look for Ed25519 key data (00 00 00 20)
        key_data_start = private_key_buffer.find(b'\x00\x00\x00\x40')
        if key_data_start == -1:
            raise Exception('Unable to find Ed25519 key data')

        # The key starts after 0x00 0x00 0x00 0x20 (32-byte key length marker)
        return private_key_buffer[key_data_start + 4: key_data_start + 68]

    @staticmethod
    def decrypt_with_private_key(ed25519_private_key: str, encrypted_data: str) -> str:
        """
        Decrypts data using an Ed25519 private key (converted to Curve25519).
        
        :param ed25519_private_key: The OpenSSH Ed25519 private key string.
        :param encrypted_data: The encrypted data in base64 format.
        :return: The decrypted plaintext as a string.
        """
        Ed25519Crypto.initialize()

        try:
            # Parse the OpenSSH private key format and extract the Ed25519 private key
            ed25519_private_key_bytes = Ed25519Crypto.parse_openssh_private_key(ed25519_private_key)

            # Convert Ed25519 private key to Curve25519 private key for use with Box
            curve25519_private_key_bytes = crypto_sign_ed25519_sk_to_curve25519(ed25519_private_key_bytes)
            private_key = PrivateKey(curve25519_private_key_bytes)

            # Decode the base64-encoded encrypted data
            encrypted_buffer = base64.b64decode(encrypted_data)

            # Extract nonce (24 bytes), ephemeral public key (32 bytes), and ciphertext
            nonce = encrypted_buffer[:24]
            ephemeral_public_key = PublicKey(encrypted_buffer[-32:])
            ciphertext = encrypted_buffer[24:-32]

            # Create a Box with the recipient's Curve25519 private key and the sender's ephemeral public key
            box = Box(private_key, ephemeral_public_key)

            # Decrypt the ciphertext
            decrypted = box.decrypt(ciphertext, nonce)

            return decrypted.decode('utf-8')

        except Exception as e:
            raise Exception("Failed to decrypt with the provided Ed25519 private key: " + str(e))




#!/usr/bin/env python
import os
from setuptools import setup, find_packages

# Dictionary to hold package metadata
about = {}
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "cngn_manager", "__version__.py"), "r", encoding="utf-8") as f:
    exec(f.read(), about)
with open(os.path.join(here, "README.md"), "r", encoding="utf-8") as f:
    readme = f.read()

# Define package dependencies
requires = [
    "requests",
    "cryptography==43.0.1",
    "pynacl==1.5.0",
    "mnemonic==0.20",
    "bip32utils==0.3.post4",
    "hashlib",
    "tronpy",
    "stellar-sdk==11.1.0",
]

# Define test dependencies
test_requirements = [
    "pytest-httpbin==2.1.0",
    "pytest-cov",
    "pytest-mock",
    "pytest-xdist",
    "PySocks>=1.5.6, !=1.5.7",
    "pytest>=3",

]

# Setup function to handle packaging
setup(
    name=about["__title__"],
    version=about["__version__"],
    description=about["__description__"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    license=about["__license__"],
    url=about["__url__"],
    long_description=readme,
    long_description_content_type="text/markdown",
    install_requires=requires,
    python_requires=">=3.8",
    packages=find_packages(),
    project_urls={
        "Documentation": about["__url__"],
        "Source": "https://github.com/wrappedcbdc/cngn-pyhon-library.git",
    }
)
